@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2009 Junjiro R. Okajima
+ * Copyright (C) 2005-2011 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,6 +25,14 @@
 
 #ifdef __KERNEL__
 
+#define AuStub(type, name, body, ...) \
+	static inline type name(__VA_ARGS__) { body; }
+
+#define AuStubVoid(name, ...) \
+	AuStub(void, name, , __VA_ARGS__)
+#define AuStubInt0(name, ...) \
+	AuStub(int, name, return 0, __VA_ARGS__)
+
 #include "debug.h"
 
 #include "branch.h"
@@ -33,6 +41,7 @@
 #include "dbgaufs.h"
 #include "dentry.h"
 #include "dir.h"
+#include "dynop.h"
 #include "file.h"
 #include "fstype.h"
 #include "inode.h"
